(* $Id: configuration.mli,v 1.8 2002/09/02 18:06:42 berdine Exp $ *)

open Types

type goal =
  | GGetPkgsPossible of coords
  | GGetPkgsDefinite of coords * Package.IdSet.t
  | GDeliverPkg of Package.id
  | GExplore

module GoalSet : sig
  type elt = goal
  and t
  val empty : t
  val is_empty : t -> bool
  val mem : elt -> t -> bool
  val add : elt -> t -> t
  val singleton : elt -> t
  val remove : elt -> t -> t
  val union : t -> t -> t
  val inter : t -> t -> t
  val diff : t -> t -> t
  val compare : t -> t -> int
  val equal : t -> t -> bool
  val subset : t -> t -> bool
  val iter : (elt -> unit) -> t -> unit
  val fold : (elt -> 'a -> 'a) -> t -> 'a -> 'a
  val for_all : (elt -> bool) -> t -> bool
  val exists : (elt -> bool) -> t -> bool
  val filter : (elt -> bool) -> t -> t
  val partition : (elt -> bool) -> t -> t * t
  val cardinal : t -> int
  val elements : t -> elt list
  val min_elt : t -> elt
  val max_elt : t -> elt
  val choose : t -> elt
end

val goals : GoalSet.t ref

val goal_leq : goal -> goal -> bool

val money : int ref

type robot_desc = {mutable posn: coords; mutable pkgs: Package.IdSet.t}
val robots : (robot_id, robot_desc) Hashtbl.t

val self : unit -> robot_desc

val pkgs_to_deliver_here : unit -> Package.IdSet.t

val input_update : in_channel -> unit
val input_packages : in_channel -> unit

val pr_config : unit -> unit
val pr_goal : goal -> unit
