(* $Id: distance.ml,v 1.23 2002/09/02 18:22:03 ccris Exp $ *)

open Types

let my_pos () = (Configuration.self ()).Configuration.posn

let my_saved_pos = ref({x = -100; y = -100})

type dir = {dn:bool; de:bool; ds:bool; dw:bool}

let distarray = ref (Array.create 0 0)


let empty = -1
let dist_sep = -2
let border = -3
let wall = -4

let rec make_divisible_by d n =
  if n mod d = 0 then n else make_divisible_by d (n+1)

let make_distarray () = 
  let w = !World.map_rep.World.width
  and h = !World.map_rep.World.height in
  let index_of_coords (x,y) = x + 1 + (w+2)*((h-1-y)+1) in
  let n = (w+2) * (h+2) in
  let a = Array.make n empty in
  for i = 0 to w-1
  do
    for j = 0 to h-1
    do
      match !World.map_rep.World.tiles.(j).[i] with
      | '#' | '~' -> a.(index_of_coords (i,j)) <- wall
      | _ -> ()
    done
  done;
  for i=0 to w+1 do a.(i) <- border done;
  for i=n-(w+2) to n-1 do a.(i) <- border done;
  for i=1 to h do a.(i*(w+2)) <- border done;
  for i=1 to h do a.(i*(w+2)+w+1) <- border done;
  let good_power_of_two = 1024 in
  let q_size = make_divisible_by good_power_of_two (2*w+2*h+1) in
  let q = Array.make q_size 0 in
  let incrq x =
    x := if !x mod good_power_of_two = good_power_of_two -1 then (!x + 1) mod q_size else !x + 1 in
  q.(0) <- index_of_coords (!my_saved_pos.x, !my_saved_pos.y);
  q.(1) <- dist_sep;
  let l = ref 0 in
  let l1 = ref 1 in
  let dist = ref 0 in
  while !l != !l1 do
   assert (!l != !l1 +1);
   let i = q.(!l) in
     (* print_current (); *)
     incrq l;
     if (i != dist_sep)
     then if a.(i) = empty
       then
         begin
           a.(i) <- !dist;
	   if a.(i-1) = empty then (incrq l1; q.(!l1) <- i-1);
	   if a.(i+1) = empty then (incrq l1; q.(!l1) <- i+1;
	      (* Format.printf "i+1:%d a.(i+1):%d" (i+1) a.(i+1);
               Format.print_newline () *));
	   if a.(i-(w+2)) = empty then (incrq l1; q.(!l1) <- i-(w+2));
	   if a.(i+(w+2)) = empty then (incrq l1; q.(!l1) <- i+(w+2))
         end
       else ()
     else
       begin
         incr dist;
         (* incr l; *)
	 incrq l1;
         q.(!l1) <- dist_sep
       end
 done;
 a

let print_arr a =
 let w = !World.map_rep.World.width
 and h = !World.map_rep.World.height in
 let a = !distarray in
 let pp n = match n with
            | 0 -> Format.printf "%3s" "*"
	    | n when n=empty -> Format.printf "%3s" "."
	    | n when n=border -> Format.printf "%3s" "+"
	    | n when n=wall -> Format.printf "%3s" "#"
	    | _ -> Format.printf "%3d" n in
 for j=0 to h+1 do
  for i=0 to w+1 do pp a.(i + (w+2)*j) done;
  Format.print_newline ()
 done

let refresh_distarray () =
  let p = my_pos ()
  in if p = !my_saved_pos then ()
     else begin
            my_saved_pos := p;
	    distarray := make_distarray ()
	    (* ;print_arr () *)
	  end

exception Origin_not_reachable

type card = DN | DE | DS | DW

let pdir d =
  let s = match d with
          | DN -> "N"
          | DE -> "E"
          | DS -> "S"
          | DW -> "W"
  in Format.printf("%s ") s

let mark_back dest =
 let w = !World.map_rep.World.width
 and h = !World.map_rep.World.height in
 let index_of_coords (x,y) = x + 1 + (w+2)*((h-1-y)+1) in
 let a = !distarray in
 let r = ref (index_of_coords (dest.x, dest.y)) in
 let dir = ref DN in
 let dist = a.(!r) in
 while a.(!r) != 0
 do
    let i = !r in
    let next = Array.of_list [(i-1,DE); (i+1,DW); (i-(w+2),DS); (i+(w+2),DN)] in
    let j = Random.int 4 in
    let (n1,n2,n3,n4) = (next.(j), next.((j+1) mod 4), next.((j+2) mod 4), next.((j+3) mod 4)) in
    let n =
      if a.(fst n1) + 1 = a.(i) then n1 else
      if a.(fst n2) + 1 = a.(i) then n2 else
      if a.(fst n3) + 1 = a.(i) then n3 else
      if a.(fst n4) + 1 = a.(i) then n4
      else raise Origin_not_reachable in
    (* Format.printf "i: %d " i; pdir d; Format.print_newline (); *)
    dir := snd n;
    r := fst n; 
    (* a.(i) <- border *)
 done;
 (!dir,dist)

let print_dir d =
  let s = match d with
          | (DN,_) -> "N"
          | (DE,_) -> "E"
          | (DS,_) -> "S"
          | (DW,_) -> "W"
  in Format.printf("The shortest distance suggests: %s") s;
     Format.print_newline ()

let print_dir d =
  let s = match d with
          | (DN,_) -> "N"
          | (DE,_) -> "E"
          | (DS,_) -> "S"
          | (DW,_) -> "W"
  in Format.printf("The shortest distance suggests: %s") s;
     Format.print_newline ()

let get_dirs dest =
  refresh_distarray ();
  let falses = {dn= false; de= false; ds= false; dw= false} in
  (try
    let d = mark_back dest
    in (* print_arr ();
       print_dir d; *)
       match d with
       | (DN,dist) -> Some ({falses with dn= true}, dist)
       | (DE,dist) -> Some ({falses with de= true}, dist)
       | (DS,dist) -> Some ({falses with ds= true}, dist)
       | (DW,dist) -> Some ({falses with dw= true}, dist)
  with Origin_not_reachable -> None)
