(* $Id: global_decision.ml,v 1.14 2002/09/02 18:54:44 berdine Exp $ *)

open Types
open Configuration

type direction = char

type priority = int (* non-zero *)
let default_prio = 1

type dir_wghts = int * int * int * int * int

type command =
	Local_decision.command =
  | CMove of priority * direction
  | CPick of priority * Package.IdSet.t
  | CDrop of priority * Package.IdSet.t

let init () =
  Random.self_init ();
  Local_decision.init ();
  for y = 0 to !World.map_rep.World.height-1 do
	for x = 0 to !World.map_rep.World.width-1 do
	  if !World.map_rep.World.tiles.(y).[x] = '@'
	  then goals := (GoalSet.add (GGetPkgsPossible {x= x; y= y}) !goals)
	done
  done

let choose_goal () =
  let g_lst = GoalSet.elements !goals in
  let sorted_g_lst = Sort.list goal_leq g_lst in
  let g = List.hd sorted_g_lst in
  pr_goal g;
  g


exception GoalPositionNotReachable

let get_dir_wghts pos =
  let module D = Distance in
  let iob b = if b then 1 else 0 in
  match D.get_dirs pos with
  | None -> raise GoalPositionNotReachable
  | Some({D.dn= d; D.de= e; D.ds= s; D.dw= w}, _) ->
	  (iob d, iob e, iob s, iob w, 0)


exception Return of Local_decision.direction_weights * command

let choose_command () =
  let rec loop () =
	try
	  match choose_goal () with
	  | GGetPkgsPossible(pos) ->
		  assert (pos <> (self()).posn);
              (* we have arrived, and know what packages are here *)
		  raise (Return ((get_dir_wghts pos),
						 CMove(1,' ')))
	  | GGetPkgsDefinite(pos, pkgs) ->
		  if pos = (self()).posn
		  then (* we have arrived *)
                (* calculate packages to Pick, possibly replace this with
				   a knapsack algorithm *)
			raise (Return ((0, 0, 0, 0, 1), CPick(1,pkgs)))
		  else (* still travelling *)
			raise (Return ((get_dir_wghts pos), CMove(1,' ')))
	  | GDeliverPkg(pkg) ->
		  let pos = (Package.desc pkg).Package.d in
		  if pos = (self()).posn
		  then (* we have arrived *)
			let pkgs = pkgs_to_deliver_here () in
			(if not (Package.IdSet.is_empty pkgs)
			 then raise (Return ((0, 0, 0, 0, 1), CDrop(1,pkgs))))
		  else (* still travelling *)
			raise (Return ((get_dir_wghts pos), CMove(1,' ')))
	  | GExplore -> ()
	  with GoalPositionNotReachable -> loop ()
  in
  let (wts, cmd) =
	try
	  begin
		loop ();
		((0, 0, 0, 0, 1), CDrop(1, Package.IdSet.empty))
	  end
	with Return(dw,cmd) -> (dw, cmd)
  in
  match Local_decision.choose_command wts, cmd with
  | CMove(_) as cm, _ -> cm
  | CDrop(p,_), CPick(_,ps) -> CPick(p,ps)
  | CDrop(p,_), CDrop(_,ps) -> CDrop(p,ps)
  | _ -> assert false


let output ochan str =
  output_string ochan (str ^ "\n");
  flush ochan;
  print_endline ("< " ^ str)

let output_command ochan =
  let f id st = (string_of_int id) ^ " " ^ st in
  match choose_command() with
  | CMove(p,d) ->
	  output ochan ((string_of_int p) ^ " Move " ^ (String.make 1 d) ^ "")
  | CPick(p,ps) ->
	  (output ochan
		 ((string_of_int p) ^ " Pick " ^ (Package.IdSet.fold f ps "")))
  | CDrop(p,ps) ->
	  (output ochan
		 ((string_of_int p) ^ " Drop " ^ (Package.IdSet.fold f ps "")))
