(* $Id: local_decision.ml,v 1.17 2002/09/02 18:50:45 pablo Exp $ *)

open Types
open Configuration
open Global_decision

type direction_weights = int * int * int * int * int

type direction = Global_decision.direction
type priority = Global_decision.priority
type command = Global_decision.command =
               | CMove of priority * direction
               | CPick of priority * Package.IdSet.t
               | CDrop of priority * Package.IdSet.t


let water, wall, pack, enemy, empty = 50, 20, 20, 20, 1

let rec gcd x y = 
  if x < y then gcd y x
  else let z = x mod y in 
  if z = 0 then y
  else gcd y z

let lcm x y = x*y/(gcd x y)

let rec lcml = function
  [] -> 1
| x::tl -> lcm x (lcml tl)

let length = 6

let weights = Array.make_matrix  (length+1) (length+1) 0

let dir = Array.make_matrix  (length+1) (length+1) (0,0,0,0,0)

let init_weights() = 
  let rec mklist = function
    0 -> [1]
  | n -> (4*n)::(mklist (n-1))
  in let wl = mklist length
  in let factor = lcml wl
(* (x/y)+c is the extra weight for the direction "forward" *)
(*  in let x,y,c = 3,2,2 *)
  in 
  for i = 0 to length / 2 do
    for j = 0 to length / 2 do
      weights.(i).(j) <- 
	  factor / (List.nth wl (j+i));
      weights.(i).(length - j) <- 
	  factor / (List.nth wl (j+i));
      weights.(length - i).(j) <- 
	  factor / (List.nth wl (j+i));
      weights.(length-i).(length - j) <- 
	  factor / (List.nth wl (j+i))
    done
  done;
  weights.(length/2).(length/2) <- weights.(length/2).(length/2-1) 

(* Directions are north, east, south, west, stay put *)

let init_dirs() = 
  for i = 0 to length / 2 - 1 do
      (* row by row *)
    for j = 0 to length / 2 - 1 do
	  let f a b = if a > b then 2 else 1 
	  in let ij, ji = f i j, f j i
	  in
      (* south west *)
	  dir.(i).(j) <- 
	  (0,0,weights.(i).(j)/ij,weights.(i).(j)/ji,0);
      (* south east *)
	  dir.(i).(length - j) <- 
	  (0,weights.(i).(length-j)/ij,weights.(i).(length-j)/ji,0,0);
      (* north west *)
	  dir.(length - i).(j) <- 
	  (weights.(length-i).(j)/ji,0,0,weights.(length-i).(j)/ij,0);
      (* north east *)
	  dir.(length-i).(length - j) <- 
	  (weights.(length-i).(length-j)/ji,
	   weights.(length-i).(length-j)/ij,0,0,0);
    done;
      (* west *)
    dir.(length/2).(i) <- 
	(0,0,0,weights.(length/2).(i),0);
      (* east *)
    dir.(length/2).(length-i) <- 
	(0,weights.(length/2).(length-i),0,0,0);
      (* south *)
    dir.(i).(length/2) <- 
	(0,0,weights.(i).(length/2),0,0);
      (* north *)
    dir.(length-i).(length/2) <- 
	(weights.(length-i).(length/2),0,0,0,0);
  done;
  dir.(length/2).(length/2) <-
      (0,0,0,0,(weights.(length/2).(length/2)/2))

let init() = 
  init_weights ();
  init_dirs ()

let water, wall, pack, enemy, empty = 10, 1, 3, 4, 1

let initial_weights = ref (0,0,0,0,0)

let add_tuples (f,r,b,l,p) (f1,r1,b1,l1,p1) n =
  (f+f1*n,r+r1*n,b+b1*n,l+l1*n,p+p1*n)

let mul_tuples (f,r,b,l,p) (f1,r1,b1,l1,p1)  =
  (f*f1,r*r1,b*b1,l*l1,p*p1)

exception Finished

let get_move array iw =
  let total = ref (0,0,0,0,0)
  in let bid = ref 1 
  in let outweight = 5
  in let la = Array.length array - 1
  in let (i1,i2,i3,i4,i5) = iw in
  initial_weights := (outweight*i1+1,outweight*i2+1,outweight*i3+1,
			 outweight*i4+1,outweight*i5+1);
  for i = 0 to la do
	let s = array.(i) 
	in let ls = String.length s - 1
	in
	for j = 0 to ls - 1 do
	  match s.[j] with
		'#' -> 
		  total := add_tuples !total dir.(la-i).(ls-j) wall
	  | '~' -> 
		total := add_tuples !total dir.(la-i).(ls-j) water
	  | 'p' ->
		total := add_tuples !total dir.(i).(j) pack
	  | '*' ->
		total := add_tuples !total dir.(i).(j) enemy;
		bid := 2
	  | 'x' -> 
		total := add_tuples !total dir.(i).(j) pack;
		total := add_tuples !total dir.(i).(j) enemy;
		bid := 2
	  | _ -> total := add_tuples !total iw empty
	done
  done;
  let (north,east,south,west,put) =  mul_tuples !total !initial_weights 
(*       in let _ = print_int north; print_string " "; *)
(* 	print_int east; print_string " "; *)
(* 	print_int south; print_string " "; *)
(* 	print_int west; print_string " "; *)
(* 	print_int put; print_string "\n"; *)
  in let lst = ref []
  in let rec add_ordered m d = function
	[] -> [(m,d)]
  | (x,e)::tl -> 
	  if m < x then (x,e)::(add_ordered m d tl)
	  else (m,d)::(x,e)::tl
  in 
  lst := add_ordered north 'N' !lst;
  lst := add_ordered east 'E' !lst;
  lst := add_ordered west 'W' !lst;
  lst := add_ordered south 'S' !lst;
  lst := add_ordered put 'p' !lst;
  let res = ref (List.hd !lst) in
  try 
	begin
	  while !lst <> [] do
		let (m,d)::tl = !lst in
		match d with 
		| 'N' -> 
			if array.(length/2).[length/2+1] == '~'
			then (lst := tl; res := List.hd !lst)
			else raise Finished
		| 'E' -> 
			if array.(length/2+1).[length/2] == '~'
			then (lst := tl; res := List.hd !lst)
			else raise Finished
		| 'S' -> 
			if array.(length/2).[length/2-1] == '~'
			then (lst := tl; res := List.hd !lst)
			else raise Finished
		| 'W' -> 
			if array.(length/2-1).[length/2] == '~'
			then (lst := tl; res := List.hd !lst)
			else raise Finished
		| _ -> raise Finished
	  done;
	  let (_,m) = !res in (!bid,m)
	end
  with Finished -> let (_,m) = !res in (!bid,m)

let a = Array.init(length+1) (fun i -> ".......")
let initial = (0,0,1,0,0)

let choose_command initial_tuple =
  let arr = Array.init (length+1) (fun i -> String.create (length+1)) 
  in let {x= x; y= y} = (self()).posn 
  in let (i0, j0, i1, j1) = (x - length/2, y - length/2,
				x + length/2, y + length/2) 
  in let convert i j = (i - i0, j - j0) 
  in let place_robots _ {posn= {x= i; y= j}; pkgs=_} =
    if i0 <= i && i <= i1 && j0 <= j && j <= j1 then
	  let (i', j') = convert i j in
	  if i' <= length && j' <= length
	  then arr.(j').[i'] <- '*'
  in let place_pkgs = function
  | GGetPkgsPossible({x= i; y= j})
  | GGetPkgsDefinite({x= i; y= j},_) ->
	  if i0 <= i && i <= i1 && j0 <= j && j <= j1 then
		let (i', j') = convert i j in
	    if i' <= length && j' <= length
	    then (match arr.(j').[i'] with
			  | '*' -> arr.(j').[i'] <- 'x'
			  | _ -> arr.(j').[i'] <- 'p')
  | _ -> ()
  in
  begin
    for i = i0 to i1 do
      for j = j0 to j1 do
        let (i', j') = convert i j in
		(arr.(j').[i'] <-
		 if (0 <= i && i < !World.map_rep.World.width
			 && 0 <= j && j < !World.map_rep.World.height)
		 then !World.map_rep.World.tiles.(j).[i]
		 else '#')
      done
    done;
    Hashtbl.iter place_robots robots;
	GoalSet.iter place_pkgs !goals;
    match get_move arr initial_tuple with
	| (b,'p') -> CDrop (b, Package.IdSet.empty)
	| (b,c) -> CMove(b, c)
  end
